// xmpp-client: A sample GTK client in Rust
// Copyright (C) 2024 Link Mauve <linkmauve@linkmauve.fr>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use adw::prelude::*;
use gtk::{gio, glib};

mod message;
mod poezio_logs;
mod tab;
mod window;

use message::Message;
use tab::Tab;

fn get_own_nick() -> String {
    std::env::var("USER").unwrap()
}

fn load_logs(jid: &str) -> gio::ListStore {
    let logs = {
        let mut file = std::fs::File::open(jid).unwrap();
        let mut string = String::new();
        use std::io::Read;
        file.read_to_string(&mut string).unwrap();
        string
    };
    let (_, logs) = poezio_logs::parse_logs(&logs).unwrap();
    let store = gio::ListStore::new::<Message>();
    for item in logs {
        if let poezio_logs::Item::Message(message) = item {
            use poezio_logs::LogItem;
            let message = Message::new(
                message.get_time(),
                message.get_nick(),
                &message.get_message(),
            );
            store.append(&message);
        }
    }
    store
}

fn xep_0392(input: &str) -> String {
    use sha1::Digest;
    let sha1 = sha1::Sha1::digest(input);
    let hue = ((sha1[1] as u16) << 8 | sha1[0] as u16) as f64 / 65536. * 360.;
    let (r, g, b) = hsluv::hsluv_to_rgb(hue, 100., 75.);
    let (r, g, b) = ((r * 255.) as u8, (g * 255.) as u8, (b * 255.) as u8);
    format!("#{:02x}{:02x}{:02x}", r, g, b)
}

fn main() {
    let app = adw::Application::builder()
        .application_id("fr.linkmauve.XmppClient")
        .flags(gio::ApplicationFlags::HANDLES_COMMAND_LINE)
        .build();

    let tabs_store = gio::ListStore::new::<Tab>();

    let tabs_store2 = tabs_store.clone();
    app.connect_command_line(move |app, command_line| {
        let args = command_line.arguments();
        let mut iter = args.iter();
        iter.next().unwrap();
        let mut first_tab = None;
        for os_arg in iter {
            let jid = os_arg.to_str().unwrap();
            let tab = Tab::new(jid, "User");
            tabs_store2.append(&tab);
            if first_tab.is_none() {
                first_tab = Some(tab);
            }
        }
        if let Some(tab) = first_tab {
            let win = app.active_window().unwrap();
            let win: &window::MainWindow = win.downcast_ref().unwrap();
            let store = load_logs(&tab.jid());
            win.selection().set_model(Some(&store));
            win.messages()
                .scroll_to(win.selection().n_items() - 1, gtk::ListScrollFlags::FOCUS, None);
        }
        0
    });

    app.connect_startup(move |app| {
        let win = window::MainWindow::new(app);

        let action_close = gio::ActionEntry::builder("close")
            .activate(|window: &window::MainWindow, _, _| {
                window.close();
            })
            .build();
        win.add_action_entries([action_close]);
        app.set_accels_for_action("win.close", &["<Ctrl>Q"]);

        assert!(Tab::static_type().is_valid());
        assert!(Message::static_type().is_valid());

        let store = gio::ListStore::new::<Message>();

        let factory = gtk::SignalListItemFactory::new();
        factory.connect_setup(move |_, list_item| {
            let label = gtk::Label::new(None);
            label.set_halign(gtk::Align::Start);
            label.set_use_markup(true);
            let list_item: &gtk::ListItem = list_item.downcast_ref().unwrap();
            list_item.set_child(Some(&label));
        });
        factory.connect_bind(move |_, list_item| {
            let list_item: &gtk::ListItem = list_item.downcast_ref().unwrap();
            let message: Message = list_item.item().and_downcast().unwrap();
            let label: gtk::Label = list_item.child().and_downcast().unwrap();

            let date = message.date();
            let date = if date.ymd() == glib::DateTime::now(&glib::TimeZone::local()).unwrap().ymd()
            {
                date.format("%H:%M:%S").unwrap()
            } else {
                date.format("%Y-%m-%d %H:%M:%S").unwrap()
            };
            let author = message.author();
            let color = xep_0392(&author);
            let body = message.body();
            let body = html_escape::encode_text(&body);
            let string = format!("{date} <span color='{color}'>{author}</span>> {body}");
            label.set_label(&string);
        });

        win.messages().set_factory(Some(&factory));
        win.selection().set_model(Some(&store));

        let tabs_factory = gtk::SignalListItemFactory::new();
        tabs_factory.connect_setup(move |_, list_item| {
            let label = gtk::Label::new(None);
            label.set_halign(gtk::Align::Start);
            let list_item: &gtk::ListItem = list_item.downcast_ref().unwrap();
            list_item.set_child(Some(&label));
        });
        tabs_factory.connect_bind(move |_, list_item| {
            let list_item: &gtk::ListItem = list_item.downcast_ref().unwrap();
            let tab: Tab = list_item.item().and_downcast().unwrap();
            let label: gtk::Label = list_item.child().and_downcast().unwrap();

            label.set_label(&tab.name());
            label.set_tooltip_text(Some(&tab.jid()));
        });

        win.tabs().set_factory(Some(&tabs_factory));
        win.tabs_selection().set_model(Some(&tabs_store));

        let win2 = win.clone();
        win.tabs_selection()
            .connect_selection_changed(move |tabs_selection, _, _| {
                let item = tabs_selection.selected_item().unwrap();
                let tab: &Tab = item.downcast_ref().unwrap();
                println!("Switching to {}", tab.jid());
                let store = load_logs(&tab.jid());
                let selection = win2.selection();
                selection.set_model(Some(&store));
                win2.messages().scroll_to(
                    selection.n_items() - 1,
                    gtk::ListScrollFlags::FOCUS,
                    None,
                );
                win2.split_view().set_show_content(true);
                win2.entry().grab_focus();
            });

        let win2 = win.clone();
        win.entry().connect_activate(move |entry| {
            let text = entry.text();
            entry.set_text("");
            println!("Send: {text}");
            let message = Message::now(&get_own_nick(), &text);
            let selection = win2.selection();
            let store = selection
                .model()
                .unwrap()
                .downcast::<gio::ListStore>()
                .unwrap();
            store.append(&message);
            win2.messages()
                .scroll_to(selection.n_items() - 1, gtk::ListScrollFlags::FOCUS, None);
        });

        win.messages()
            .scroll_to(win.selection().n_items() - 1, gtk::ListScrollFlags::FOCUS, None);

        win.present();
    });

    app.run();
}
